@@ -61,27 +61,6 @@ bool isDescendantOf(const QList<ModelIndex> &list, const QModelIndex &idx)
     return false;
 }
 
-static bool isDescendantOf(const QModelIndex &ancestor, const QModelIndex &descendant)
-{
-    if (!descendant.isValid()) {
-        return false;
-    }
-
-    if (ancestor == descendant) {
-        return false;
-    }
-
-    QModelIndex parent = descendant.parent();
-    while (parent.isValid()) {
-        if (parent == ancestor) {
-            return true;
-        }
-
-        parent = parent.parent();
-    }
-    return false;
-}
-
 static bool isDescendantOf(const QItemSelection &selection, const QModelIndex &descendant)
 {
     if (!descendant.isValid()) {
@@ -436,6 +415,7 @@ public:
           m_includeAllSelected(false),
           m_rowsInserted(false),
           m_rowsRemoved(false),
+          m_recreateFirstChildMappingOnRemoval(false),
           m_rowsMoved(false),
           m_resetting(false),
           m_sourceModelResetting(false),
@@ -583,13 +563,15 @@ public:
     void sourceDataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight);
 
     void removeSelectionFromProxy(const QItemSelection &selection);
-    void removeRangeFromProxy(const QItemSelectionRange &range);
 
     void selectionChanged(const QItemSelection &selected, const QItemSelection &deselected);
     void sourceModelDestroyed();
 
     void resetInternalData();
 
+    bool rootWillBeRemoved(const QItemSelection &selection,
+                           const QModelIndex &root);
+
     /**
       When items are inserted or removed in the m_startWithChildTrees configuration,
       this method helps find the startRow for use emitting the signals from the proxy.
@@ -618,6 +600,7 @@ public:
     bool m_includeAllSelected;
     bool m_rowsInserted;
     bool m_rowsRemoved;
+    bool m_recreateFirstChildMappingOnRemoval;
     QPair<int, int> m_proxyRemoveRows;
     bool m_rowsMoved;
     bool m_resetting;
@@ -1106,93 +1089,96 @@ void KSelectionProxyModelPrivate::sourceRowsInserted(const QModelIndex &parent,
     m_pendingSelectionChanges.clear();
 }
 
+static bool rootWillBeRemovedFrom(const QModelIndex &ancestor, int start, int end,
+                               const QModelIndex &root)
+{
+  Q_ASSERT(root.isValid());
+
+  auto parent = root;
+  while (parent.isValid()) {
+      auto prev = parent;
+      parent = parent.parent();
+      if (parent == ancestor) {
+          return (prev.row() <= end && prev.row() >= start);
+      }
+  }
+  return false;
+}
+
+bool KSelectionProxyModelPrivate::rootWillBeRemoved(const QItemSelection &selection,
+                               const QModelIndex &root)
+{
+    Q_ASSERT(root.isValid());
+
+    for (auto r : selection) {
+        if (m_includeAllSelected) {
+            if (r.parent() == root.parent() && root.row() <= r.bottom() && root.row() >= r.top()) {
+                return true;
+            }
+        } else {
+            if (rootWillBeRemovedFrom(r.parent(), r.top(), r.bottom(), root)) {
+                return true;
+            }
+        }
+    }
+    return false;
+}
+
 QPair<int, int> KSelectionProxyModelPrivate::beginRemoveRows(const QModelIndex &parent, int start, int end) const
 {
     Q_Q(const KSelectionProxyModel);
 
-    QPair<int, int> pair = qMakePair(start, end);
-
-    if (m_omitChildren && !m_startWithChildTrees && !m_includeAllSelected) {
-        // SubTreeRoots
-        if (m_rootIndexList.contains(parent) || isDescendantOf(m_rootIndexList, parent)) {
-            return qMakePair(-1, -1);
-        }
-    }
-
-    const QModelIndex proxyParent = mapParentFromSource(parent);
-
     if (!m_includeAllSelected && !m_omitChildren) {
         // SubTrees and SubTreesWithoutRoots
+        const QModelIndex proxyParent = mapParentFromSource(parent);
         if (proxyParent.isValid()) {
-            return pair;
-        }
-        if (m_startWithChildTrees && m_rootIndexList.contains(parent)) {
-            // SubTreesWithoutRoots topLevel
-            const int proxyStartRow = getProxyInitialRow(parent) + start;
-            return qMakePair(proxyStartRow, proxyStartRow + (end - start));
+            return qMakePair(start, end);
         }
     }
 
-    if (m_includeAllSelected && m_startWithChildTrees) {
-        // ChildrenOfExactSelection
-        int position = m_rootIndexList.indexOf(parent);
-        if (position != -1) {
-            const int proxyStartRow = getProxyInitialRow(parent) + start;
-            int proxyEndRow = proxyStartRow + (end - start);
-            ++position;
-            while (m_rootIndexList.size() < position) {
-                const QModelIndex idx = m_rootIndexList.at(position);
-                if (isDescendantOf(parent, idx)) {
-                    proxyEndRow += q->sourceModel()->rowCount(idx);
-                } else {
-                    break;
-                }
-            }
-            return qMakePair(proxyStartRow, proxyEndRow);
-        }
-        return qMakePair(-1, -1);
+    if (m_startWithChildTrees && m_rootIndexList.contains(parent)) {
+        const int proxyStartRow = getProxyInitialRow(parent) + start;
+        const int proxyEndRow = proxyStartRow + (end - start);
+        return qMakePair(proxyStartRow, proxyEndRow);
     }
 
     QList<QPersistentModelIndex>::const_iterator rootIt = m_rootIndexList.constBegin();
     const QList<QPersistentModelIndex>::const_iterator rootEnd = m_rootIndexList.constEnd();
-    int rootPosition = 0;
-    int rootStartRemove = -1;
-    int rootEndRemove = -1;
-    int siblingCount = 0;
+    int proxyStartRemove = 0;
 
-    for (; rootIt != rootEnd; ++rootIt, ++rootPosition) {
-        if (m_omitChildren && m_includeAllSelected) {
-            // ExactSelection
-            if (parent == rootIt->parent() && rootIt->row() <= end && rootIt->row() >= start) {
-                if (rootStartRemove == -1) {
-                    rootStartRemove = rootPosition;
-                }
-                ++rootEndRemove;
-            } else {
-                if (rootStartRemove != -1) {
-                    break;
-                }
-            }
+    for (; rootIt != rootEnd; ++rootIt) {
+        if (rootWillBeRemovedFrom(parent, start, end, *rootIt)) {
+            break;
         } else {
-            if (isDescendantOf(parent, *rootIt)) {
-                if (rootStartRemove == -1) {
-                    rootStartRemove = rootPosition;
-                }
-                ++rootEndRemove;
-                if (m_startWithChildTrees) {
-                    siblingCount += q->sourceModel()->rowCount(*rootIt);
-                }
+            if (m_startWithChildTrees) {
+                proxyStartRemove += q->sourceModel()->rowCount(*rootIt);
             } else {
-                if (rootStartRemove != -1) {
-                    break;
-                }
+                ++proxyStartRemove;
             }
         }
     }
-    if (rootStartRemove != -1) {
-        return qMakePair(siblingCount + rootStartRemove, siblingCount + rootEndRemove);
+
+    if (rootIt == rootEnd) {
+        return qMakePair(-1, -1);
     }
 
+    int proxyEndRemove = proxyStartRemove;
+
+    for (; rootIt != rootEnd; ++rootIt) {
+        if (!rootWillBeRemovedFrom(parent, start, end, *rootIt)) {
+            break;
+        }
+        if (m_startWithChildTrees) {
+            proxyEndRemove += q->sourceModel()->rowCount(*rootIt);
+        } else {
+            ++proxyEndRemove;
+        }
+    }
+
+    --proxyEndRemove;
+    if (proxyEndRemove >= 0) {
+        return qMakePair(proxyStartRemove, proxyEndRemove);
+    }
     return qMakePair(-1, -1);
 }
 
@@ -1215,6 +1201,7 @@ void KSelectionProxyModelPrivate::sourceRowsAboutToBeRemoved(const QModelIndex &
 
     m_rowsRemoved = true;
     m_proxyRemoveRows = pair;
+    m_recreateFirstChildMappingOnRemoval = m_mappedFirstChildren.leftContains(q->sourceModel()->index(start, 0, parent));
     q->beginRemoveRows(proxyParent, pair.first, pair.second);
 }
 
@@ -1265,7 +1252,9 @@ void KSelectionProxyModelPrivate::endRemoveRows(const QModelIndex &sourceParent,
         // If D and F are selected, the proxy contains E, F, G, H. If B is then deleted E to H will
         // be removed, including both first child mappings at E and G.
 
-        removeFirstChildMappings(proxyStart, proxyEnd);
+        if (!proxyParent.isValid()) {
+            removeFirstChildMappings(proxyStart, proxyEnd);
+        }
     }
 
     if (proxyParent.isValid()) {
@@ -1288,10 +1277,11 @@ void KSelectionProxyModelPrivate::sourceRowsRemoved(const QModelIndex &parent, i
 {
     Q_Q(KSelectionProxyModel);
     Q_UNUSED(end)
+    Q_UNUSED(start)
 
     Q_ASSERT(parent.isValid() ? parent.model() == q->sourceModel() : true);
 
-    if (!m_selectionModel || !m_selectionModel->hasSelection()) {
+    if (!m_selectionModel) {
         return;
     }
 
@@ -1303,12 +1293,13 @@ void KSelectionProxyModelPrivate::sourceRowsRemoved(const QModelIndex &parent, i
     Q_ASSERT(m_proxyRemoveRows.first >= 0);
     Q_ASSERT(m_proxyRemoveRows.second >= 0);
     endRemoveRows(parent, m_proxyRemoveRows.first, m_proxyRemoveRows.second);
-    if (m_startWithChildTrees && start == 0 && q->sourceModel()->hasChildren(parent))
+    if (m_recreateFirstChildMappingOnRemoval && q->sourceModel()->hasChildren(parent))
         // The private endRemoveRows call might remove the first child mapping for parent, so
         // we create it again in that case.
     {
         createFirstChildMapping(parent, m_proxyRemoveRows.first);
     }
+    m_recreateFirstChildMappingOnRemoval = false;
 
     m_proxyRemoveRows = qMakePair(-1, -1);
     q->endRemoveRows();
@@ -1667,12 +1658,6 @@ void KSelectionProxyModelPrivate::removeParentMappings(const QModelIndex &parent
     SourceProxyIndexMapping::right_iterator it = m_mappedParents.rightBegin();
     SourceProxyIndexMapping::right_iterator endIt = m_mappedParents.rightEnd();
 
-    typedef QPair<QModelIndex, QPersistentModelIndex> Pair;
-
-    QList<Pair> pairs;
-
-    QModelIndexList list;
-
     const bool flatList = isFlat();
 
     while (it != endIt) {
@@ -1726,131 +1711,62 @@ void KSelectionProxyModelPrivate::removeSelectionFromProxy(const QItemSelection
         return;
     }
 
-    q->rootSelectionAboutToBeRemoved(selection);
+    QList<QPersistentModelIndex>::iterator rootIt = m_rootIndexList.begin();
+    const QList<QPersistentModelIndex>::iterator rootEnd = m_rootIndexList.end();
+    int proxyStartRemove = 0;
 
-    Q_FOREACH (const QItemSelectionRange &range, selection) {
-        removeRangeFromProxy(range);
+    for (; rootIt != rootEnd; ++rootIt) {
+        if (rootWillBeRemoved(selection, *rootIt)) {
+            break;
+        } else {
+            if (m_startWithChildTrees) {
+                auto rc = q->sourceModel()->rowCount(*rootIt);
+                proxyStartRemove += rc;
+            } else {
+                ++proxyStartRemove;
+            }
+        }
     }
-}
 
-void KSelectionProxyModelPrivate::removeRangeFromProxy(const QItemSelectionRange &range)
-{
-    Q_Q(KSelectionProxyModel);
+    if (rootIt == rootEnd) {
+        return;
+    }
 
-    Q_ASSERT(range.model() == q->sourceModel());
+    int proxyEndRemove = proxyStartRemove;
 
-    const QModelIndex sourceTopLeft = range.topLeft();
-    const QModelIndex proxyTopLeft = mapFromSource(sourceTopLeft);
-    const QModelIndex sourceBottomLeft = range.bottomRight().sibling(range.bottom(), 0);
-    const QModelIndex proxyParent = proxyTopLeft.parent();
-    const QModelIndex sourceParent = sourceTopLeft.parent();
+    QList<QPersistentModelIndex>::iterator rootRemoveStart = rootIt;
 
-    if (m_startWithChildTrees) {
-        Q_ASSERT(sourceTopLeft.isValid());
-        Q_ASSERT(sourceBottomLeft.isValid());
-        const int startRootIdx = m_rootIndexList.indexOf(sourceTopLeft);
-        int endRootIdx = m_rootIndexList.indexOf(sourceBottomLeft);
-        QItemSelection extraRanges;
-        if (m_includeAllSelected) {
-            // It can happen that indexes of descendants get in between indexes which make up a range.
-            // We handle only the first contiguous block here and handle the rest later.
-            int idx = startRootIdx;
-            const int bottomIdx = endRootIdx;
-            const int rootListSize = m_rootIndexList.size();
-            int next = idx + 1;
-            while (next <= bottomIdx) {
-                if (next < rootListSize && m_rootIndexList.at(next).parent() == sourceParent) {
-                    idx = next;
-                    ++next;
-                } else {
-                    break;
-                }
-            }
-            endRootIdx = idx;
-            ++idx;
-            while (idx <= bottomIdx) {
-                const QModelIndex index = m_rootIndexList.at(idx);
-                if (m_rootIndexList.at(idx).parent() == sourceParent) {
-                    extraRanges << QItemSelectionRange(index, index);
-                }
-                ++idx;
-            }
+    for (; rootIt != rootEnd; ++rootIt) {
+        if (!rootWillBeRemoved(selection, *rootIt)) {
+            break;
         }
-        Q_ASSERT(endRootIdx != -1);
-        int childrenCount = q->sourceModel()->rowCount(sourceTopLeft);
-        for (int rootIdx = startRootIdx + 1; rootIdx <= endRootIdx; ++rootIdx) {
-            childrenCount += q->sourceModel()->rowCount(m_rootIndexList.at(rootIdx));
+        q->rootIndexAboutToBeRemoved(*rootIt);
+        if (m_startWithChildTrees) {
+            auto rc = q->sourceModel()->rowCount(*rootIt);
+            proxyEndRemove += rc;
+        } else {
+            ++proxyEndRemove;
         }
-        if (childrenCount == 0) {
-            for (int rootIdx = startRootIdx; rootIdx <= endRootIdx; --endRootIdx) {
-                const QModelIndex idx = m_rootIndexList.at(rootIdx);
-                q->rootIndexAboutToBeRemoved(idx);
-                m_rootIndexList.removeOne(idx);
-            }
-            return;
-        }
-        if (!m_includeAllSelected) {
-            ++endRootIdx;
-            for (; endRootIdx < m_rootIndexList.size(); ++endRootIdx) {
-                const QModelIndex idx = m_rootIndexList.at(endRootIdx);
-                if (isDescendantOf(sourceBottomLeft, idx)) {
-                    childrenCount += q->sourceModel()->rowCount(idx);
-                } else {
-                    break;
-                }
-            }
-            --endRootIdx;
-        }
-        const int proxyStart = getTargetRow(startRootIdx);
-        int proxyEnd = proxyStart + childrenCount - 1;
-        q->beginRemoveRows(QModelIndex(), proxyStart, proxyEnd);
+    }
 
-        for (int rootIdx = startRootIdx; rootIdx <= endRootIdx; ++rootIdx) {
-            q->rootIndexAboutToBeRemoved(m_rootIndexList.at(rootIdx));
-        }
+    --proxyEndRemove;
+    if (proxyEndRemove >= 0) {
+        q->beginRemoveRows(QModelIndex(), proxyStartRemove, proxyEndRemove);
 
-        removeParentMappings(QModelIndex(), proxyStart, proxyEnd);
-        removeFirstChildMappings(proxyStart, proxyEnd);
-        int numRemovedChildren = 0;
-        for (int rootIdx = startRootIdx; rootIdx <= endRootIdx; --endRootIdx) {
-            const QModelIndex idx = m_rootIndexList.at(rootIdx);
-            const int childCount = q->sourceModel()->rowCount(idx);
-            m_rootIndexList.removeAt(rootIdx);
-            numRemovedChildren += childCount;
+        rootIt = m_rootIndexList.erase(rootRemoveStart, rootIt);
+
+        removeParentMappings(QModelIndex(), proxyStartRemove, proxyEndRemove);
+        if (m_startWithChildTrees) {
+            removeFirstChildMappings(proxyStartRemove, proxyEndRemove);
         }
-        updateInternalTopIndexes(proxyEnd + 1, -1 * numRemovedChildren);
+        updateInternalTopIndexes(proxyEndRemove + 1, -1 * (proxyEndRemove - proxyStartRemove + 1));
+
         q->endRemoveRows();
-        if (m_includeAllSelected) {
-            removeSelectionFromProxy(kNormalizeSelection(extraRanges));
-        }
     } else {
-        if (!proxyTopLeft.isValid()) {
-            return;
-        }
-        const int height = range.height();
-        q->beginRemoveRows(proxyParent, proxyTopLeft.row(), proxyTopLeft.row() + height - 1);
-
-        // TODO: Do this conditionally if the signal is connected to anything.
-        for (int i = 0; i < height; ++i) {
-            const QModelIndex idx = sourceTopLeft.sibling(range.top() + i, sourceTopLeft.column());
-            q->rootIndexAboutToBeRemoved(idx);
-        }
-
-        removeParentMappings(proxyParent, proxyTopLeft.row(), proxyTopLeft.row() + height - 1);
-        updateInternalIndexes(proxyParent, proxyTopLeft.row() + height, -1 * height);
-
-        for (int i = 0; i < height; ++i) {
-            const QModelIndex idx = sourceTopLeft.sibling(range.top() + i, sourceTopLeft.column());
-            Q_ASSERT(idx.isValid());
-            const bool b = m_rootIndexList.removeOne(idx);
-            Q_UNUSED(b)
-            if (!b) {
-                qDebug() << idx;
-            }
-            Q_ASSERT(b);
-        }
-
-        q->endRemoveRows();
+        rootIt = m_rootIndexList.erase(rootRemoveStart, rootIt);
+    }
+    if (rootIt != rootEnd) {
+        removeSelectionFromProxy(selection);
     }
 }
 
@@ -2187,7 +2103,7 @@ void KSelectionProxyModel::setSourceModel(QAbstractItemModel *_sourceModel)
     beginResetModel();
     d->m_resetting = true;
 
-    static const char *modelSignals[] = {
+    static const char *const modelSignals[] = {
         SIGNAL(rowsAboutToBeInserted(QModelIndex,int,int)),
         SIGNAL(rowsInserted(QModelIndex,int,int)),
         SIGNAL(rowsAboutToBeRemoved(QModelIndex,int,int)),
@@ -2201,7 +2117,7 @@ void KSelectionProxyModel::setSourceModel(QAbstractItemModel *_sourceModel)
         SIGNAL(layoutChanged()),
         SIGNAL(destroyed())
     };
-    static const char *proxySlots[] = {
+    static const char *const proxySlots[] = {
         SLOT(sourceRowsAboutToBeInserted(QModelIndex,int,int)),
         SLOT(sourceRowsInserted(QModelIndex,int,int)),
         SLOT(sourceRowsAboutToBeRemoved(QModelIndex,int,int)),
@@ -2491,10 +2407,8 @@ void KSelectionProxyModel::setSelectionModel(QItemSelectionModel *itemSelectionM
                 }
                 endResetModel();
             };
-#if QT_VERSION >= QT_VERSION_CHECK(5, 5, 0)
             connect(d->m_selectionModel.data(), &QItemSelectionModel::modelChanged,
                     this, handleSelectionModelModel);
-#endif
             handleSelectionModelModel();
         }
 
